def _format_nested_invoice(invoice):
    invoice.pop("Prepayments", None)
    invoice.pop("Payments", None)
    invoice.pop("CreditNotes", None)
    invoice.pop("Overpayments", None)


def _format_allocations(allocations):
    for allocation in allocations:
        invoice = allocation.get("Invoice", {})
        _format_nested_invoice(invoice)


def format_credit_notes(credit_notes):
    for credit_note in credit_notes:
        credit_note.pop("Payments", None)
        _format_allocations(credit_note.get("Allocations", []))


def format_contact_groups(contact_groups):
    for contact_group in contact_groups:
        contact_group.pop("Contacts", None)


def format_payments(payments):
    for payment in payments:
        invoice = payment.get("Invoice", {})
        _format_nested_invoice(invoice)


def format_over_pre_payments(over_pre_payments):
    for payment in over_pre_payments:
        _format_allocations(payment.get("Allocations", []))


def strip_warnings(records):
    for record in records:
        record.pop("Warnings", None)

format_users = strip_warnings


def format_receipts(receipts):
    strip_warnings(receipts)
    for receipt in receipts:
        receipt.get("User", {}).pop("Warnings", None)
        receipt.get("Contact", {}).pop("Warnings", None)


def format_contacts(contacts):
    strip_warnings(contacts)
    for contact in contacts:
        format_contact_groups(contact["ContactGroups"])

def format_invoices(invoices):
    # NB: Xero sometimes formats the Date as '/Date(0+0000)/' to indicate
    # it is 0 milliseconds from the unix epoch. Convert this to a datetime
    # that will be accepted by the transformer. This should not cause
    # inconsitencies because the 'Date' is normally returned as an iso8601
    # string and this edge case causes it to be returned differently
    for invoice in invoices:
        if invoice.get('Date') == '/Date(0+0000)/':
            invoice['Date'] = '1970-01-01T00:00:00.000000Z'


def format_journals(journals):
    # NB: Xero sometimes formats the JournalDate as '/Date(0+0000)/' to
    # indicate it is 0 milliseconds from the unix epoch. Convert this to a
    # datetime that will be accepted by the transformer. This should not
    # cause inconsitencies because the 'Date' is normally returned as an
    # iso8601 string and this edge case causes it to be returned
    # differently
    for journal in journals:
        if journal.get('JournalDate') == '/Date(0+0000)/':
            journal['JournalDate'] = '1970-01-01T00:00:00.000000Z'



from base64 import b64encode
import re
import json
import decimal
import sys
import math
from os.path import join
from datetime import datetime, date, time, timedelta
import requests
from singer.utils import strftime, strptime_to_utc
import six
import pytz
import backoff
import singer

LOGGER = singer.get_logger()

BASE_URL = "https://api.xero.com/api.xro/2.0"


class XeroError(Exception):
    def __init__(self, message=None, response=None):
        super().__init__(message)
        self.message = message
        self.response = response


class XeroBadRequestError(XeroError):
    pass


class XeroUnauthorizedError(XeroError):
    pass


class XeroForbiddenError(XeroError):
    pass


class XeroNotFoundError(XeroError):
    pass


class XeroPreConditionFailedError(XeroError):
    pass


class XeroTooManyError(XeroError):
    pass


class XeroTooManyInMinuteError(XeroError):
    pass


class XeroInternalError(XeroError):
    pass


class XeroNotImplementedError(XeroError):
    pass


class XeroNotAvailableError(XeroError):
    pass


ERROR_CODE_EXCEPTION_MAPPING = {
    400: {
        "raise_exception": XeroBadRequestError,
        "message": "A validation exception has occurred."
    },
    401: {
        "raise_exception": XeroUnauthorizedError,
        "message": "Invalid authorization credentials."
    },
    403: {
        "raise_exception": XeroForbiddenError,
        "message": "User doesn't have permission to access the resource."
    },
    404: {
        "raise_exception": XeroNotFoundError,
        "message": "The resource you have specified cannot be found."
    },
    412: {
        "raise_exception": XeroPreConditionFailedError,
        "message": "One or more conditions given in the request header fields were invalid."
    },
    429: {
        "raise_exception": XeroTooManyError,
        "message": "The API rate limit for your organisation/application pairing has been exceeded"
    },
    500: {
        "raise_exception": XeroInternalError,
        "message": "An unhandled error with the Xero API. Contact the Xero API team if problems persist."
    },
    501: {
        "raise_exception": XeroNotImplementedError,
        "message": "The method you have called has not been implemented."
    },
    503: {
        "raise_exception": XeroNotAvailableError,
        "message": "API service is currently unavailable."
    }
}


def parse_date(value):
    # Xero datetimes can be .NET JSON date strings which look like
    # "/Date(1419937200000+0000)/"
    # https://developer.xero.com/documentation/api/requests-and-responses
    pattern = r'Date\((\-?\d+)([-+])?(\d+)?\)'
    match = re.search(pattern, value)

    iso8601pattern = r'((\d{4})-([0-2]\d)-0?([0-3]\d)T([0-5]\d):([0-5]\d):([0-6]\d))'

    if not match:
        iso8601match = re.search(iso8601pattern, value)
        if iso8601match:
            try:
                return strptime_to_utc(value)
            except Exception:
                return None
        else:
            return None

    millis_timestamp, offset_sign, offset = match.groups()
    if offset:
        if offset_sign == '+':
            offset_sign = 1
        else:
            offset_sign = -1
        offset_hours = offset_sign * int(offset[:2])
        offset_minutes = offset_sign * int(offset[2:])
    else:
        offset_hours = 0
        offset_minutes = 0

    return datetime.utcfromtimestamp((int(millis_timestamp) / 1000)) \
           + timedelta(hours=offset_hours, minutes=offset_minutes)


def _json_load_object_hook(_dict):
    """Hook for json.parse(...) to parse Xero date formats."""
    # This was taken from the pyxero library and modified
    # to format the dates according to RFC3339
    for key, value in _dict.items():
        if isinstance(value, six.string_types):
            value = parse_date(value)
            if value:
                # NB> Pylint disabled because, regardless of idioms, this is more explicit than isinstance.
                if type(value) is date:  # pylint: disable=unidiomatic-typecheck
                    value = datetime.combine(value, time.min)
                value = value.replace(tzinfo=pytz.UTC)
                _dict[key] = strftime(value)
    return _dict


def update_config_file(config, config_path):
    with open(config_path, 'w') as config_file:
        json.dump(config, config_file, indent=2)


def is_not_status_code_fn(status_code):
    def gen_fn(exc):
        if getattr(exc, 'response', None) and getattr(exc.response, 'status_code',
                                                      None) and exc.response.status_code not in status_code:
            return True
        # Retry other errors up to the max
        return False

    return gen_fn


def retry_after_wait_gen():
    while True:
        # This is called in an except block so we can retrieve the exception
        # and check it.
        exc_info = sys.exc_info()
        resp = exc_info[1].response
        sleep_time_str = resp.headers.get('Retry-After')
        LOGGER.info("API rate limit exceeded -- sleeping for %s seconds", sleep_time_str)
        yield math.floor(float(sleep_time_str))


class XeroClient():
    def __init__(self, config):
        self.session = requests.Session()
        self.user_agent = config.get("user_agent")
        self.tenant_id = None
        self.access_token = None

    def refresh_credentials(self, config, config_path):

        header_token = b64encode((config["client_id"] + ":" + config["client_secret"]).encode('utf-8'))

        headers = {
            "Authorization": "Basic " + header_token.decode('utf-8'),
            "Content-Type": "application/x-www-form-urlencoded"
        }

        post_body = {
            "grant_type": "refresh_token",
            "refresh_token": config["refresh_token"],
        }
        resp = self.session.post("https://identity.xero.com/connect/token", headers=headers, data=post_body)

        if resp.status_code != 200:
            raise_for_error(resp)
        else:
            resp = resp.json()

            # Write to config file
            config['refresh_token'] = resp["refresh_token"]
            update_config_file(config, config_path)
            self.access_token = resp["access_token"]
            self.tenant_id = config['tenant_id']

    @backoff.on_exception(backoff.expo, (json.decoder.JSONDecodeError, XeroInternalError), max_tries=3)
    @backoff.on_exception(retry_after_wait_gen, XeroTooManyInMinuteError, giveup=is_not_status_code_fn([429]),
                          jitter=None, max_tries=3)
    def check_platform_access(self, config, config_path):

        # Validating the authentication of the provided configuration
        self.refresh_credentials(config, config_path)

        headers = {
            "Authorization": "Bearer " + self.access_token,
            "Xero-Tenant-Id": self.tenant_id,
            "Content-Type": "application/json"
        }

        # Validating the authorization of the provided configuration
        contacts_url = join(BASE_URL, "Employees")
        request = requests.Request("GET", contacts_url, headers=headers)
        response = self.session.send(request.prepare())

        if response.status_code != 200:
            raise_for_error(response)

    @backoff.on_exception(backoff.expo, (json.decoder.JSONDecodeError, XeroInternalError), max_tries=3)
    @backoff.on_exception(retry_after_wait_gen, XeroTooManyInMinuteError, giveup=is_not_status_code_fn([429]),
                          jitter=None, max_tries=3)
    def filter(self, tap_stream_id, since=None, **params):
        xero_resource_name = tap_stream_id.title().replace("_", "").replace("-", "/")
        url = join(BASE_URL, xero_resource_name)
        headers = {"Accept": "application/json",
                   "Authorization": "Bearer " + self.access_token,
                   "Xero-tenant-id": self.tenant_id}
        if self.user_agent:
            headers["User-Agent"] = self.user_agent
        if since:
            headers["If-Modified-Since"] = since

        request = requests.Request("GET", url, headers=headers, params=params)
        response = self.session.send(request.prepare())
        if response.status_code != 200:
            raise_for_error(response)
            return None
        else:
            response_meta = json.loads(response.text,
                                       object_hook=_json_load_object_hook,
                                       parse_float=decimal.Decimal)
            if "/" in xero_resource_name:
                resource_name, _ = xero_resource_name.split('/')
                response_body = response_meta.pop(resource_name)
            else:
                response_body = response_meta.pop(xero_resource_name)
            return response_body


def raise_for_error(resp):
    try:
        resp.raise_for_status()
    except (requests.HTTPError, requests.ConnectionError) as error:
        try:
            error_code = resp.status_code

            # Handling status code 429 specially since the required information is present in the headers
            if error_code == 429:
                resp_headers = resp.headers
                api_rate_limit_message = ERROR_CODE_EXCEPTION_MAPPING[429]["message"]
                message = "HTTP-error-code: 429, Error: {}. Please retry after {} seconds".format(
                    api_rate_limit_message, resp_headers.get("Retry-After"))

                # Raise XeroTooManyInMinuteError exception if minute limit is reached
                if resp_headers.get("X-Rate-Limit-Problem") == 'minute':
                    raise XeroTooManyInMinuteError(message, resp) from None
            # Handling status code 403 specially since response of API does not contain enough information
            elif error_code in (403, 401):
                api_message = ERROR_CODE_EXCEPTION_MAPPING[error_code]["message"]
                message = "HTTP-error-code: {}, Error: {}".format(error_code, api_message)
            else:
                # Forming a response message for raising custom exception
                try:
                    response_json = resp.json()
                except Exception:
                    response_json = {}

                message = "HTTP-error-code: {}, Error: {}".format(
                    error_code,
                    response_json.get(
                        "error", response_json.get(
                            "Title", response_json.get(
                                "Detail", ERROR_CODE_EXCEPTION_MAPPING.get(
                                    error_code, {}).get("message", "Unknown Error")
                            ))))

            exc = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {}).get("raise_exception", XeroError)
            raise exc(message, resp) from None

        except (ValueError, TypeError):
            raise XeroError(error) from None



#!/usr/bin/env python3
import os
import json
import singer
from singer import metadata, metrics, utils
from singer.catalog import Catalog, CatalogEntry, Schema
from . import streams as streams_
from .client import raise_for_error, XeroClient, BASE_URL
from .context import Context


REQUIRED_CONFIG_KEYS = [
    "start_date",
    "client_id",
    "client_secret",
    "tenant_id",
    "refresh_token",

]

LOGGER = singer.get_logger()

BAD_CREDS_MESSAGE = (
    "Failed to refresh OAuth token using the credentials from both the config and S3. "
    "The token might need to be reauthorized from the integration's properties "
    "or there could be another authentication issue. Please attempt to reauthorize "
    "the integration."
)


class BadCredsException(Exception):
    pass


def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def load_schema(tap_stream_id):
    path = "schemas/{}.json".format(tap_stream_id)
    schema = utils.load_json(get_abs_path(path))
    dependencies = schema.pop("tap_schema_dependencies", [])
    refs = {}
    for sub_stream_id in dependencies:
        refs[sub_stream_id] = load_schema(sub_stream_id)
    if refs:
        singer.resolve_schema_references(schema, refs)
    return schema

def load_metadata(stream, schema):
    mdata = metadata.new()

    mdata = metadata.write(mdata, (), 'table-key-properties', stream.pk_fields)
    mdata = metadata.write(mdata, (), 'forced-replication-method', stream.replication_method)

    if stream.bookmark_key:
        mdata = metadata.write(mdata, (), 'valid-replication-keys', [stream.bookmark_key])

    for field_name in schema['properties'].keys():
        if field_name in stream.pk_fields or field_name == stream.bookmark_key:
            mdata = metadata.write(mdata, ('properties', field_name), 'inclusion', 'automatic')
        else:
            mdata = metadata.write(mdata, ('properties', field_name), 'inclusion', 'available')

    return metadata.to_list(mdata)


def ensure_credentials_are_valid(config):
    XeroClient(config).filter("currencies")


def discover(ctx):
    ctx.check_platform_access()
    catalog = Catalog([])
    for stream in streams_.all_streams:
        schema_dict = load_schema(stream.tap_stream_id)
        mdata = load_metadata(stream, schema_dict)
        schema = Schema.from_dict(schema_dict)
        catalog.streams.append(CatalogEntry(
            stream=stream.tap_stream_id,
            tap_stream_id=stream.tap_stream_id,
            key_properties=stream.pk_fields,
            schema=schema,
            metadata=mdata
        ))
    return catalog


def load_and_write_schema(stream):
    singer.write_schema(
        stream.tap_stream_id,
        load_schema(stream.tap_stream_id),
        stream.pk_fields,
    )


def sync(ctx):
    ctx.refresh_credentials()
    currently_syncing = ctx.state.get("currently_syncing")
    start_idx = streams_.all_stream_ids.index(currently_syncing) \
        if currently_syncing else 0
    stream_ids_to_sync = [cs.tap_stream_id for cs in ctx.catalog.streams
                          if cs.is_selected()]
    streams = [s for s in streams_.all_streams[start_idx:]
               if s.tap_stream_id in stream_ids_to_sync]
    for stream in streams:
        ctx.state["currently_syncing"] = stream.tap_stream_id
        ctx.write_state()
        load_and_write_schema(stream)
        LOGGER.info("Syncing stream: %s", stream.tap_stream_id)
        stream.sync(ctx)
    ctx.state["currently_syncing"] = None
    ctx.write_state()



def main_impl():
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)
    if args.discover:
        discover(Context(args.config, {}, {}, args.config_path)).dump()
        print()
    else:
        if args.catalog:
            catalog = args.catalog
        else:
            LOGGER.info("Running sync without provided Catalog. Discovering.")
            catalog = discover(Context(args.config, {}, {}, args.config_path))

        sync(Context(args.config, args.state, catalog, args.config_path))

def main():
    try:
        main_impl()
    except Exception as exc:
        LOGGER.critical(exc)
        raise exc


if __name__ == "__main__":
    main()



import singer
from singer import bookmarks as bks_
from .client import XeroClient


class Context():
    def __init__(self, config, state, catalog, config_path):
        self.config = config
        self.config_path = config_path
        self.state = state
        self.catalog = catalog
        self.client = XeroClient(config)

    def refresh_credentials(self):
        self.client.refresh_credentials(self.config, self.config_path)

    def check_platform_access(self):
        self.client.check_platform_access(self.config, self.config_path)

    def get_bookmark(self, path):
        return bks_.get_bookmark(self.state, *path)

    def set_bookmark(self, path, val):
        bks_.write_bookmark(self.state, path[0], path[1], val)

    def get_offset(self, path):
        off = bks_.get_offset(self.state, path[0])
        return (off or {}).get(path[1])

    def set_offset(self, path, val):
        bks_.set_offset(self.state, path[0], path[1], val)

    def clear_offsets(self, tap_stream_id):
        bks_.clear_offset(self.state, tap_stream_id)

    def update_start_date_bookmark(self, path):
        val = self.get_bookmark(path)
        if not val:
            val = self.config["start_date"]
            self.set_bookmark(path, val)
        return val

    def write_state(self):
        singer.write_state(self.state)



from requests.exceptions import HTTPError
import singer
from singer import metadata, metrics, Transformer
from singer.utils import strptime_with_tz
import backoff
from . import transform

import datetime
from dataclasses import dataclass

from dateutil.relativedelta import relativedelta

LOGGER = singer.get_logger()
FULL_PAGE_SIZE = 100
MAX_PERIODS_ALLOWED = 11
LOOKUP_YEARS = 4


def _request_with_timer(tap_stream_id, xero, filter_options):
    with metrics.http_request_timer(tap_stream_id) as timer:
        try:
            resp = xero.filter(tap_stream_id, **filter_options)
            timer.tags[metrics.Tag.http_status_code] = 200
            return resp
        except HTTPError as e:
            timer.tags[metrics.Tag.http_status_code] = e.response.status_code
            raise


class RateLimitException(Exception):
    pass


@backoff.on_exception(backoff.expo,
                      RateLimitException,
                      max_tries=10,
                      factor=2)
def _make_request(ctx, tap_stream_id, filter_options=None, attempts=0):
    filter_options = filter_options or {}
    try:
        return _request_with_timer(tap_stream_id, ctx.client, filter_options)
    except HTTPError as e:
        if e.response.status_code == 401:
            if attempts == 1:
                raise Exception("Received Not Authorized response after credential refresh.") from e
            ctx.refresh_credentials()
            return _make_request(ctx, tap_stream_id, filter_options, attempts + 1)

        if e.response.status_code == 503:
            raise RateLimitException() from e

        raise
    assert False


class Stream():
    def __init__(self, tap_stream_id, pk_fields, bookmark_key="UpdatedDateUTC", format_fn=None):
        self.tap_stream_id = tap_stream_id
        self.pk_fields = pk_fields
        self.format_fn = format_fn or (lambda x: x)
        self.bookmark_key = bookmark_key
        self.replication_method = "INCREMENTAL"
        self.filter_options = {}

    def metrics(self, records):
        with metrics.record_counter(self.tap_stream_id) as counter:
            counter.increment(len(records))

    def write_records(self, records, ctx):
        stream = ctx.catalog.get_stream(self.tap_stream_id)
        schema = stream.schema.to_dict()
        mdata = stream.metadata
        for rec in records:
            with Transformer() as transformer:
                rec = transformer.transform(rec, schema, metadata.to_map(mdata))
                singer.write_record(self.tap_stream_id, rec)
        self.metrics(records)


class BookmarkedStream(Stream):
    def sync(self, ctx):
        bookmark = [self.tap_stream_id, self.bookmark_key]
        start = ctx.update_start_date_bookmark(bookmark)
        records = _make_request(ctx, self.tap_stream_id, dict(since=start))
        if records:
            self.format_fn(records)
            self.write_records(records, ctx)
            max_bookmark_value = max([record[self.bookmark_key] for record in records])
            ctx.set_bookmark(bookmark, max_bookmark_value)
            ctx.write_state()


class PaginatedStream(Stream):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def sync(self, ctx):
        bookmark = [self.tap_stream_id, self.bookmark_key]
        offset = [self.tap_stream_id, "page"]
        start = ctx.update_start_date_bookmark(bookmark)
        curr_page_num = ctx.get_offset(offset) or 1

        self.filter_options.update(dict(since=start, order="UpdatedDateUTC ASC"))

        max_updated = start
        while True:
            ctx.set_offset(offset, curr_page_num)
            ctx.write_state()
            self.filter_options["page"] = curr_page_num
            records = _make_request(ctx, self.tap_stream_id, self.filter_options)
            if records:
                self.format_fn(records)
                self.write_records(records, ctx)
                max_updated = records[-1][self.bookmark_key]
            if not records or len(records) < FULL_PAGE_SIZE:
                break
            curr_page_num += 1
        ctx.clear_offsets(self.tap_stream_id)
        ctx.set_bookmark(bookmark, max_updated)
        ctx.write_state()


class BalanceSheetReportStream(Stream):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def sync(self, ctx):
        bookmark = [self.tap_stream_id, self.bookmark_key]
        offset = [self.tap_stream_id, "page"]
        start = ctx.update_start_date_bookmark(bookmark)
        curr_page_num = ctx.get_offset(offset) or 1
        max_updated = start
        while True:
            ctx.set_offset(offset, curr_page_num)
            ctx.write_state()
            date = datetime.date.today() - relativedelta(years=curr_page_num - 1)
            self.filter_options.update(dict(date=date, timeframe="MONTH", periods=MAX_PERIODS_ALLOWED))
            records = _make_request(ctx, self.tap_stream_id, self.filter_options)
            is_finished = curr_page_num >= LOOKUP_YEARS
            if not is_finished:
                self.format_fn(records)
                self.write_records(records, ctx)
                max_updated = records[-1][self.bookmark_key]
            else:
                break
            curr_page_num += 1

        ctx.clear_offsets(self.tap_stream_id)
        ctx.set_bookmark(bookmark, max_updated)
        ctx.write_state()


class ProfitAndLossStream(Stream):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def sync(self, ctx):
        bookmark = [self.tap_stream_id, self.bookmark_key]
        offset = [self.tap_stream_id, "page"]
        start = ctx.update_start_date_bookmark(bookmark)
        curr_page_num = ctx.get_offset(offset) or 1
        max_updated = start
        while True:
            ctx.set_offset(offset, curr_page_num)
            ctx.write_state()
            from_date = self.first_day_of_month(datetime.date.today()) - relativedelta(months=curr_page_num)
            to_date = self.last_day_of_month(from_date)
            self.filter_options.update(dict(fromDate=from_date, toDate=to_date))
            records = _make_request(ctx, self.tap_stream_id, self.filter_options)
            is_finished = curr_page_num > LOOKUP_YEARS * 12
            if not is_finished:
                self.format_fn(records)
                self.write_records(records, ctx)
                max_updated = records[-1][self.bookmark_key]
            else:
                break
            curr_page_num += 1

        ctx.clear_offsets(self.tap_stream_id)
        ctx.set_bookmark(bookmark, max_updated)
        ctx.write_state()

    @staticmethod
    def last_day_of_month(any_day):
        next_month = any_day.replace(day=28) + datetime.timedelta(days=4)  # this will never fail
        return next_month - datetime.timedelta(days=next_month.day)

    @staticmethod
    def first_day_of_month(any_day):
        return datetime.date(any_day.year, any_day.month, 1)


class Contacts(PaginatedStream):
    def __init__(self, *args, **kwargs):
        super().__init__("contacts", ["ContactID"], format_fn=transform.format_contacts, *args, **kwargs)

    def sync(self, ctx):
        # Parameter to collect archived contacts from the Xero platform
        if ctx.config.get("include_archived_contacts") in ["true", True]:
            self.filter_options.update({'includeArchived': "true"})

        super().sync(ctx)


class Journals(Stream):
    """The Journals endpoint is a special case. It has its own way of ordering
    and paging the data. See
    https://developer.xero.com/documentation/api/journals"""

    def sync(self, ctx):
        bookmark = [self.tap_stream_id, self.bookmark_key]
        journal_number = ctx.get_bookmark(bookmark) or 0
        while True:
            filter_options = {"offset": journal_number}
            records = _make_request(ctx, self.tap_stream_id, filter_options)
            if records:
                self.format_fn(records)
                self.write_records(records, ctx)
                journal_number = max((record[self.bookmark_key] for record in records))
                ctx.set_bookmark(bookmark, journal_number)
                ctx.write_state()
            if not records or len(records) < FULL_PAGE_SIZE:
                break


class LinkedTransactions(Stream):
    """The Linked Transactions endpoint is a special case. It supports
    pagination, but not the Modified At header, but the objects returned have
    the UpdatedDateUTC timestamp in them. Therefore we must always iterate over
    all of the data, but we can manually omit records based on the
    UpdatedDateUTC property."""

    def sync(self, ctx):
        bookmark = [self.tap_stream_id, self.bookmark_key]
        offset = [self.tap_stream_id, "page"]
        start = ctx.update_start_date_bookmark(bookmark)
        curr_page_num = ctx.get_offset(offset) or 1
        max_updated = start
        while True:
            ctx.set_offset(offset, curr_page_num)
            ctx.write_state()
            filter_options = {"page": curr_page_num}
            raw_records = _make_request(ctx, self.tap_stream_id, filter_options)
            records = [x for x in raw_records
                       if strptime_with_tz(x[self.bookmark_key]) >= strptime_with_tz(start)]
            if records:
                self.write_records(records, ctx)
                max_updated = records[-1][self.bookmark_key]
            if not records or len(records) < FULL_PAGE_SIZE:
                break
            curr_page_num += 1
        ctx.clear_offsets(self.tap_stream_id)
        ctx.set_bookmark(bookmark, max_updated)
        ctx.write_state()


class Everything(Stream):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bookmark_key = None
        self.replication_method = "FULL_TABLE"

    def sync(self, ctx):
        records = _make_request(ctx, self.tap_stream_id)
        self.format_fn(records)
        self.write_records(records, ctx)


all_streams = [
    # PAGINATED STREAMS
    # These endpoints have all the best properties: they return the
    # UpdatedDateUTC property and support the Modified After, order, and page
    # parameters
    PaginatedStream("bank_transactions", ["BankTransactionID"]),
    Contacts(),
    PaginatedStream("quotes", ["QuoteID"]),
    PaginatedStream("credit_notes", ["CreditNoteID"], format_fn=transform.format_credit_notes),
    PaginatedStream("invoices", ["InvoiceID"], format_fn=transform.format_invoices),
    PaginatedStream("manual_journals", ["ManualJournalID"]),
    PaginatedStream("overpayments", ["OverpaymentID"], format_fn=transform.format_over_pre_payments),
    PaginatedStream("payments", ["PaymentID"], format_fn=transform.format_payments),
    PaginatedStream("prepayments", ["PrepaymentID"], format_fn=transform.format_over_pre_payments),
    PaginatedStream("purchase_orders", ["PurchaseOrderID"]),

    # JOURNALS STREAM
    # This endpoint is paginated, but in its own special snowflake way.
    Journals("journals", ["JournalID"], bookmark_key="JournalNumber", format_fn=transform.format_journals),

    # NON-PAGINATED STREAMS
    # These endpoints do not support pagination, but do support the Modified At
    # header.
    BookmarkedStream("accounts", ["AccountID"]),
    BookmarkedStream("bank_transfers", ["BankTransferID"], bookmark_key="CreatedDateUTC"),
    BookmarkedStream("employees", ["EmployeeID"]),
    BookmarkedStream("expense_claims", ["ExpenseClaimID"]),
    BookmarkedStream("items", ["ItemID"]),
    BookmarkedStream("receipts", ["ReceiptID"], format_fn=transform.format_receipts),
    BookmarkedStream("users", ["UserID"], format_fn=transform.format_users),

    # PULL EVERYTHING STREAMS
    # These endpoints do not support the Modified After header (or paging), so
    # we must pull all the data each time.
    Everything("branding_themes", ["BrandingThemeID"]),
    Everything("contact_groups", ["ContactGroupID"], format_fn=transform.format_contact_groups),
    Everything("currencies", ["Code"]),
    Everything("organisations", ["OrganisationID"]),
    Everything("repeating_invoices", ["RepeatingInvoiceID"]),
    Everything("tax_rates", ["TaxType"]),
    Everything("tracking_categories", ["TrackingCategoryID"]),

    # LINKED TRANSACTIONS STREAM
    # This endpoint is not paginated, but can do some manual filtering
    LinkedTransactions("linked_transactions", ["LinkedTransactionID"], bookmark_key="UpdatedDateUTC"),

    # REPORTS STREAM
    # This endpoint is not paginated, but can do some manual filtering
    BalanceSheetReportStream("reports-balance_sheet", ["ReportID"], bookmark_key="UpdatedDateUTC"),
    ProfitAndLossStream("reports-profit_and_loss", ["ReportID"], bookmark_key="UpdatedDateUTC"),

]
all_stream_ids = [s.tap_stream_id for s in all_streams]
